#include <stdio.h>

/*********************from linux kernel include elf.h*********************************/
typedef unsigned long   __u32;
typedef unsigned short  __u16;
typedef long            __s32;

/* 32-bit ELF base types. */
typedef __u32   Elf32_Addr;
typedef __u16   Elf32_Half;
typedef __u32   Elf32_Off;
typedef __s32   Elf32_Sword;
typedef __u32   Elf32_Word;


#define EI_NIDENT       16

typedef struct elf32_hdr{
  unsigned char e_ident[EI_NIDENT];
  Elf32_Half    e_type;
  Elf32_Half    e_machine;
  Elf32_Word    e_version;
  Elf32_Addr    e_entry;  /* Entry point */
  Elf32_Off     e_phoff;
  Elf32_Off     e_shoff;
  Elf32_Word    e_flags;
  Elf32_Half    e_ehsize;
  Elf32_Half    e_phentsize;
  Elf32_Half    e_phnum;
  Elf32_Half    e_shentsize;
  Elf32_Half    e_shnum;
  Elf32_Half    e_shstrndx;
} Elf32_Ehdr;

/* These constants define the permissions on sections in the program
   header, p_flags. */
#define PF_R		0x4
#define PF_W		0x2
#define PF_X		0x1

typedef struct elf32_phdr{
  Elf32_Word	p_type;
  Elf32_Off	p_offset;
  Elf32_Addr	p_vaddr;
  Elf32_Addr	p_paddr;
  Elf32_Word	p_filesz;
  Elf32_Word	p_memsz;
  Elf32_Word	p_flags;
  Elf32_Word	p_align;
} Elf32_Phdr;

typedef struct elf32_shdr{
  Elf32_Word	sh_name;
  Elf32_Word	sh_type;
  Elf32_Word	sh_flags;
  Elf32_Addr	sh_addr;
  Elf32_Off	sh_offset;
  Elf32_Word	sh_size;
  Elf32_Word	sh_link;
  Elf32_Word	sh_info;
  Elf32_Word	sh_addralign;
  Elf32_Word	sh_entsize;
} Elf32_Shdr;


typedef struct elf32_sym{
  Elf32_Word    st_name;
  Elf32_Addr    st_value;
  Elf32_Word    st_size;
  unsigned char st_info;
  unsigned char st_other;
  Elf32_Half    st_shndx;
} Elf32_Sym;

/******************************end linux kernel elf.h********************************/


/************************************* to dump symtab*****************************/
int  symtab_offset = 0; 
int  symtab_size   = 0;
char * symtab_buf;

int  strtab_offset = 0;
int  strtab_size   = 0;
char * strtab_buf;

void usage ( char * file )
{

	printf(" usage: %s <elf filename> \n",file);
	exit (0) ;


}

int
elf_fread (char * file_buf,int count,
	   int block, FILE *fd)
{
	int ret;
		
	ret = fread ( file_buf , count ,block ,fd);

	
	if (ret == 0 ) printf("read file error\n");
	
	else return 1;
	
}

void dump_head_info ( Elf32_Ehdr * elf_head)
{
	/*class*/
	switch (elf_head->e_ident[4]){
       
	case 1:
		printf (" 32-bits ");
		break;
	case 2:
		printf (" 64-bits ");
		break ;

	default:
		printf (" error bit");
	        break ;
	}

	/*data type end big or little*/
	switch (elf_head->e_ident[5]){
       
	case 1:
		printf (" little-end ");
		break;
	case 2:
		printf (" big-end");
		break ;

	default:
		printf (" error end");
	        break ;
	}

	/*elf file type*/
	switch (elf_head->e_type){
       
	case 1:
		printf (" relocatable file ");
		break;
	case 2:
		printf (" executable file ");
		break ;
	case 3:
		printf (" shared object file ");
		break ;
	case 4:
		printf (" core file ");
		break ;

	default:
		printf (" unknow file type");
	        break ;
	}

	/*machine type */
	switch (elf_head->e_machine){
       
	case 1:
		printf (" AT&T ");
		break;
	case 2:
		printf (" SPARC ");
		break ;
	case 3:
		printf (" Intel 80x86 ");
		break ;
	case 4:
		printf (" Motolora 68k ");
		break ;
	case 5:
		printf (" Motolora 88k ");
		break ;
	case 7:
		printf (" Intel 80860 ");
		break ;
	case 8:
		printf (" MIPS BLX(China) ");
		break ;

	default:
		printf (" unknow machine type " );
	        break ;
	}

	printf ( "!\n " );

	printf ("*-----------------dump other info--------------------*\n");

	printf ("  entry : 0x%08X \n",elf_head->e_entry );
	
}


void dump_phead_info (Elf32_Phdr * elf_phead)
{

	printf("Phdr type : %d offset : 0x%08x vaddr : 0x%08x \n",
			elf_phead->p_type,
			elf_phead->p_offset,
			elf_phead->p_vaddr);


}

void dump_shead_info (Elf32_Shdr * elf_shead ,char * shstrtab_buf)
{

	printf ( "Shdr file offset: 0x%08x sh_name: %s \n", 
			elf_shead->sh_offset,
			shstrtab_buf +elf_shead->sh_name);

	if(strncmp(shstrtab_buf + elf_shead->sh_name, ".symtab",7) ==0 ){
	       	symtab_offset = elf_shead->sh_offset;
		symtab_size   = elf_shead->sh_size;
	}
	if(strncmp(shstrtab_buf + elf_shead->sh_name, ".strtab",7) ==0 ){
	       	strtab_offset = elf_shead->sh_offset;
		strtab_size   = elf_shead->sh_size ;
	}

}

/****************************************************
 *
 * return 0 failed ,
 * return sh_offset is shstrtab start
 *
 *
 ****************************************************
 * */

int find_shstrtab (Elf32_Shdr * elf_shead,int e_shoff)
{

	long temp;
	
		
/* -to debug- */	
//	printf(" offset 0x%08x size: 0x%08x \n",elf_shead->sh_offset ,elf_shead->sh_size);
	temp = elf_shead->sh_offset + elf_shead->sh_size ;
	
	/*e_shoff is 4 byte start !*/
	if (temp & 3 ) {
	   temp = ((temp + 4) & 0xfffffffc); 
	}
	
	if(temp  == e_shoff) return elf_shead->sh_offset ;
	else return 0;
	
}

void dump_symtab_info (void)
{
	int size;
	int i;
	Elf32_Sym *elf_symbol;
	
	size = sizeof (struct elf32_sym);
	for (i=0 ; i<symtab_size / size ; i++){
	
		elf_symbol = (Elf32_Sym *) (symtab_buf + size * i); 
		printf ("symbol val: 0x%08x name: %s\n",
				elf_symbol->st_value,
				strtab_buf +elf_symbol->st_name );
	}
}

void dump_info ( char *filename )

{

        FILE *fd;
	char file_buf [1024];
	char * p_buf, * s_buf , * temp_buf, *all_section_buf;
	char c;
	int  hdr_size;
	int  all_section_offset;
	int  temp_size;
	
	int  shstrtab_start;
	char * shstrtab_buf;


	
	Elf32_Ehdr elf_hdr;
	Elf32_Ehdr *elf_head;
	Elf32_Phdr *elf_phead;
	Elf32_Shdr *elf_shead;
	int i;
	
	
	hdr_size = sizeof (struct elf32_hdr);
	
	fd = fopen ( filename , "r" );
	if (fd == NULL){
		printf("open %s error!\n",filename);
		exit (0);
	}
		
	
	elf_fread (file_buf , 1 , hdr_size , fd);

	if ( (c  = strncmp (file_buf , "\117ELF", 4)) < 0) {
		printf(" \"%s\" is not elf file\n",filename);
		exit (0);
	}


	memcpy ( &elf_hdr, file_buf , hdr_size);
	
	elf_head  = (Elf32_Ehdr * ) (&elf_hdr);

	dump_head_info (elf_head);

	/*-----------------read form hdr_size to e_phoff size----------------*/
	if (elf_head->e_phoff - hdr_size)	elf_fread (file_buf , 1 , elf_head->e_phoff - hdr_size , fd);


	/*--------------program head info-----------------------*/

	p_buf = (char *)malloc ( elf_head->e_phnum * elf_head->e_phentsize );
	
	elf_fread (p_buf , 1, elf_head->e_phnum * elf_head->e_phentsize,fd);
	
	for (i = 0 ;i < elf_head -> e_phnum  ; i++) {

	
		elf_phead = (Elf32_Phdr *) (p_buf + i * elf_head->e_phentsize);
					     

		dump_phead_info (elf_phead);

	}

	/*------------------read from p_head to s_head size, them are all sections-------*/
	all_section_offset = elf_head->e_phnum * elf_head->e_phentsize + elf_head->e_phoff ; 
	temp_size          = elf_head->e_shoff - all_section_offset;
	
	all_section_buf    = (char *) malloc(temp_size);
	elf_fread (all_section_buf, 1 ,temp_size, fd);
	
	/*-------------------read section head table  info----------------*/
	/*----------if file too big size,this is a error----------------- */
	
	temp_size = elf_head->e_shnum * elf_head->e_shentsize;
	s_buf     = (char *)malloc (temp_size);
	
	elf_fread (s_buf, 1 ,temp_size, fd);
	
	/*-------------------segment info-------------------- */
	printf ( "Segment info  (%d) offset 0x%08x \n",elf_head->e_shnum,elf_head->e_shoff );
	for (i = 0 ;i < elf_head -> e_shnum  ; i++) {
		
	
		elf_shead = (Elf32_Shdr *) (s_buf + i* elf_head->e_shentsize);
  		if( (shstrtab_start  =	find_shstrtab (elf_shead, elf_head->e_shoff) ) != 0 ) break;
		else shstrtab_start  = 0;

	}


	if (shstrtab_start !=0){
		shstrtab_buf = all_section_buf + (shstrtab_start - all_section_offset);
		
		for (i = 0 ;i < elf_head -> e_shnum  ; i++) {
			elf_shead = (Elf32_Shdr *) (s_buf + i* elf_head->e_shentsize);
			dump_shead_info (elf_shead,shstrtab_buf);
		}
	}
	

	/*--------------------dump symbol table info----------------*/
	
	
	if ( symtab_offset && strtab_offset ) {
	/*------------------- read from s_head to symbol size----------*/
	        if (!lseek (fd, symtab_offset,SEEK_SET)){ 
			printf("symbol SEEK error!\n");
			return ;
		}
		symtab_buf = (char *) malloc (symtab_size);
		strtab_buf = (char *) malloc (strtab_size);
		elf_fread  (symtab_buf , 1,symtab_size,fd);

		
	        if (!lseek (fd, strtab_offset,SEEK_SET)){ 
			printf("strtab SEEK error!\n");
			return ;
		}
		
		elf_fread  (strtab_buf , 1,strtab_size,fd);
		
		dump_symtab_info ();		
	
		
	}else printf ("no symbol table\n");

		
		

	
	
	fclose (fd);
}

int main (int argv, char ** argc)
{

	
	if ( argv < 2 )
		usage (  argc[0] );

	dump_info ( argc [1] );

}
